import pyglet
from pyglet.window import key
from pyglet.gl import *

from typing import List, Dict, Tuple, Optional

from .resources.asset_definitions import *

FONT_NAME = ('Verdana', 'Helvetica', 'Arial')
FONT_SIZE = 36
ARENA_WIDTH = 640
ARENA_HEIGHT = 480

OVERLAY_SPLASH = "splash"
OVERLAY_LEADERBOARD = None
OVERLAY_GAME_OVER = "game_over"
OVERLAY_MAIN_MENU = "main_menu"
OVERLAY_INSTRUCTIONS = "instruction"
OVERLAY_PAUSED = "paused"
OVERLAY_INGAME_SETTINGS = "in_game_settings"
OVERLAY_OPTIONS = "options"
OVERLAY_NONE = None

CMD_QUIT = "quit"
CMD_START = "start"
CMD_CONTINUE = "continue"
CMD_LOAD_MAIN = "load_main"
CMD_LOAD_INSTRUCTIONS = "load_instructions"
CMD_LOAD_OPTIONS = "load_options"
CMD_NONE = None

TEXT_INSTRUCTIONS = "Play the game. Get good"
TEXT_INSTRUCTIONS_MOVE = "Use [Left] or [Right] arrow keys to move."
TEXT_INSTRUCTIONS_FIRE = "On a carriage, press [Space] to fire weapon."


class Overlay:
    def update(self, dt):
        pass

    def draw(self):
        pass

    def get_state(self):
        return CMD_NONE


class OverlayManager:
    def __init__(self, window: pyglet.window.Window, batch=None):
        self.window: pyglet.window.Window = window
        self.size = window.get_size()
        self.overlay = None
        self.state = None

        self.overlays = {OVERLAY_SPLASH: Banner("Cowboys vs Pirates",
                                                timeout=1,
                                                dismiss_func=lambda inst: self.set_overlay(OVERLAY_MAIN_MENU),
                                                window_size=self.size,
                                                text="",
                                                # batch=self.batch
                                                ),
                         OVERLAY_MAIN_MENU: MainMenu(window_size=self.size),  # , batch=self.batch),
                         OVERLAY_INSTRUCTIONS: PauseMenu("Instructions", window_size=self.size),
                         OVERLAY_PAUSED: PauseMenu("Paused", window_size=self.size),
                         #OVERLAY_OPTIONS: None,
                         OVERLAY_INGAME_SETTINGS: InGameMenu(window_size=self.size),
                         OVERLAY_GAME_OVER: Banner(
                            "Game Over",
                            dismiss_func=lambda dt: self.set_overlay(OVERLAY_MAIN_MENU),
                            window_size=self.size,
                            text="Press [Space] to try again"
                         ),
                         OVERLAY_LEADERBOARD: None,
                         }

        # pause - instructions
        # escape - in-game setting menu (quit, options, retry)
        #
        # splash - timed
        # start menu
        #     instructions (scrolling, back)
        #     options (sfx toggle, music toggle, music volume)
        # game-over
        #     scoreboard (continue)
        #     (quit, retry, back to start menu)

    def draw(self):
        self.handle_states()

        if self.overlay:
            self.overlay.draw()

    def set_overlay(self, name: Optional[str]):
        # new_overlay = self.overlays.get(name)
        if self.overlay is not None:
            self.window.remove_handlers(self.overlay)
        self.overlay = self.overlays.get(name)
        if self.overlay is not None:
            self.window.push_handlers(self.overlay)
        self.window.clear()

    def is_displaying(self):
        return self.overlay is not None

    def handle_states(self):

        # print(self.overlay)
        if not self.overlay:
            return

        state = self.overlay.get_state()

        if state == self.state:
            return

        if state == CMD_QUIT:
            pyglet.app.exit()
        elif state == CMD_START:
            self.set_overlay(OVERLAY_NONE)
        elif state == CMD_CONTINUE:
            self.set_overlay(OVERLAY_NONE)
        elif state == CMD_LOAD_MAIN:
            self.set_overlay(OVERLAY_MAIN_MENU)
        elif state == CMD_LOAD_INSTRUCTIONS:
            self.set_overlay(OVERLAY_INSTRUCTIONS)
        elif state == CMD_LOAD_OPTIONS:
            self.set_overlay(OVERLAY_OPTIONS)
        # else:
        #    raise NotImplementedError("State not specified")
        self.state = state


class Banner(Overlay):
    # FONT_SIZE = 36
    POSITION = (50, 50)
    OPACITY = 150
    COLOR = (0, 0, 0)

    TEXT_OFFSET = 100

    def __init__(self, label, timeout=None, dismiss_func=None, text="", window_size=(200, 200)):  # , batch=None):
        self.batch = pyglet.graphics.Batch()
        backdrop_size = tuple(dimension - 2 * self.POSITION[i] for i, dimension in enumerate(window_size))
        self.backdrop = pyglet.shapes.Rectangle(*self.POSITION,
                                                *backdrop_size,
                                                color=self.COLOR,
                                                batch=self.batch)
        self.backdrop.opacity = self.OPACITY
        self.text = pyglet.text.Label(label,
                                      font_name=FONT_NAME,
                                      font_size=FONT_SIZE,
                                      x=window_size[0] // 2,
                                      y=window_size[1] * 2 // 3,
                                      anchor_x='center',
                                      anchor_y='center',
                                      batch=self.batch
                                      )

        self.text = pyglet.text.Label(text,
                                      font_name=FONT_NAME,
                                      font_size=FONT_SIZE,
                                      x=window_size[0] // 2,
                                      y=window_size[1] * 2 // 3 - self.TEXT_OFFSET,
                                      anchor_x='center',
                                      anchor_y='center',
                                      batch=self.batch
                                      )

        self.dismiss_func = dismiss_func
        self.timeout = timeout
        if timeout and dismiss_func:
            pyglet.clock.schedule_once(dismiss_func, timeout)

    def draw(self):
        self.batch.draw()

    def on_key_press(self, symbol, modifiers):
        # print(symbol)

        exit_banner = False
        if symbol in (key.ESCAPE, key.ENTER, key.SPACE):
            exit_banner = True
            # print(symbol)

        if (self.dismiss_func and not self.timeout) or exit_banner:
            self.dismiss_func(0)
        return True


class Menu(Overlay):
    DEFAULT_FONT_SIZE = 36
    DEFAULT_LOCATION = (ARENA_WIDTH // 2, 350)

    ITEM_OFFSET = 60
    ITEM_SPACING = 40
    POSITION = (50, 50)
    BACKDROP_OPACITY = 128
    BACKDROP_COLOR = (0, 0, 0)

    def __init__(self, title, window_size=(200, 200), batch=None):
        self.items: List[MenuItem] = []
        self.batch = batch

        self.menu_position = (window_size[0] // 2, window_size[1] * 3 // 4)
        self.item_offset_y = self.menu_position[1] - self.ITEM_OFFSET

        backdrop_size = tuple(dimension - 2 * self.POSITION[i] for i, dimension in enumerate(window_size))
        self.backdrop = pyglet.shapes.Rectangle(*self.POSITION,
                                                *backdrop_size,
                                                color=self.BACKDROP_COLOR,
                                                batch=self.batch)
        self.backdrop.opacity = self.BACKDROP_OPACITY
        self.title_text = pyglet.text.Label(title,
                                            font_name=FONT_NAME,
                                            font_size=self.DEFAULT_FONT_SIZE,
                                            x=self.menu_position[0],
                                            y=self.menu_position[1],
                                            anchor_x='center',
                                            anchor_y='center',
                                            batch=self.batch,
                                            )

    def reset(self):
        self.state = None
        self.selected_index = 0
        self.items[self.selected_index].selected = True

    def on_key_press(self, symbol, modifiers):
        # print("DSDF")
        if symbol == key.DOWN:
            self.selected_index += 1
        elif symbol == key.UP:
            self.selected_index -= 1
        self.selected_index = min(max(self.selected_index, 0),
                                  len(self.items) - 1)
        # if symbol in (key.DOWN, key.UP) and enable_sound:
        #     bullet_sound.play()

    def on_key_release(self, symbol, modifiers):
        self.items[self.selected_index].on_key_release(symbol, modifiers)

    def draw(self):
        self.title_text.draw()
        for i, item in enumerate(self.items):
            item.draw(i == self.selected_index)

    def add_items(self, items: Dict[int, Tuple[str, str]], batch=None):
        for i in range(0, len(items)):
            label, state = items[i]
            position_x = self.menu_position[0]
            position_y = self.item_offset_y - self.ITEM_SPACING * (len(items) - 1 + i)
            self.items.append(MenuItem(label=label,
                                       x=position_x,
                                       y=position_y,
                                       activate_func=self.set_state,
                                       activate_args=state,
                                       batch=batch))
            # print(label, state, position_y)

    def set_state(self, state: str):
        self.state = state

    def get_state(self):
        return self.state


class MenuItem:
    POINTER_COLOR = (.46, 0, 1.)
    inverted_pointers = False

    DEFAULT_FONT_SIZE = 14
    DEFAULT_POSITION_X = ARENA_WIDTH // 2

    def __init__(self, label, y, activate_func, activate_args, x=DEFAULT_POSITION_X, batch=None):
        self.y = y
        self.text = pyglet.text.Label(label,
                                      font_name=FONT_NAME,
                                      font_size=self.DEFAULT_FONT_SIZE,
                                      x=x,
                                      y=y,
                                      anchor_x='center',
                                      anchor_y='center',
                                      batch=batch)
        self.activate_func = activate_func
        self.activate_args = activate_args
        self.selected = False

    def activate(self):
        self.activate_func(self.activate_args)

    def draw_pointer(self, x, y, flip=False):
        # Tint the pointer image to a color
        glPushAttrib(GL_CURRENT_BIT)
        glColor3f(*self.POINTER_COLOR)
        if flip:
            POINTER_IMG_FLIP.blit(x, y)
        else:
            POINTER_IMG.blit(x, y)
        glPopAttrib()

    def draw(self, selected):
        self.text.draw()

        if selected:
            self.draw_pointer(
                self.text.x - self.text.content_width // 2 -
                POINTER_IMG.width // 2,
                self.y,
                self.inverted_pointers)
            self.draw_pointer(
                self.text.x + self.text.content_width // 2 +
                POINTER_IMG.width // 2,
                self.y,
                not self.inverted_pointers)

    def on_key_release(self, symbol, modifiers):
        if symbol == key.ENTER and self.activate_func:
            self.activate()
            # if enable_sound:
            #     bullet_sound.play()


class ToggleMenuItem(MenuItem):
    POINTER_COLOR = (.27, .82, .25)
    inverted_pointers = True

    def __init__(self, label, value, y, toggle_func):
        self.value = value
        self.label = label
        self.toggle_func = toggle_func
        super(ToggleMenuItem, self).__init__(self.get_label(), y, None)

    def get_label(self):
        return self.label + (self.value and ': ON' or ': OFF')

    def on_key_release(self, symbol, modifiers):
        if symbol == key.LEFT or symbol == key.RIGHT:
            self.value = not self.value
            self.text.text = self.get_label()
            self.toggle_func(self.value)
            # if enable_sound:
            #     bullet_sound.play()


class MainMenu(Menu):

    def __init__(self, window_size=(200, 200), batch=None):
        self.batch = pyglet.graphics.Batch()

        self.text = pyglet.text.Label('Press [Enter] to Select',
                                      font_name=FONT_NAME,
                                      font_size=12,
                                      x=window_size[0] // 2,
                                      y=window_size[1] * 2 // 3,
                                      anchor_x='center',
                                      anchor_y='center',
                                      batch=self.batch
                                      )

        super(MainMenu, self).__init__('Cowboys VS Pirates', window_size=window_size, batch=self.batch)

        buttons = {0: ("Start Game", CMD_START),
                   1: ("Start with Instructions", CMD_LOAD_INSTRUCTIONS),
                   # 2: ("Options", CMD_LOAD_OPTIONS),
                   2: ("Quit", CMD_QUIT)}
        self.add_items(buttons, self.batch)

        # self.items.append(MenuItem('New Game', 240, lambda: self.set_state("start")))
        # self.items.append(MenuItem('Instructions', 200, lambda: self.set_state("load_instructions")))
        # self.items.append(MenuItem('Options', 160, lambda: self.set_state("load_options")))
        # self.items.append(MenuItem('Quit', 120, lambda: self.set_state("quit")))
        self.reset()

    def draw(self):
        # FIXME fix this draw thing. Something with hierarchy
        super().draw()
        self.batch.draw()


class PauseMenu(Menu):
    TEXT_OFFSET = FONT_SIZE * 1.5
    FONT_SIZE = 24

    def __init__(self, label, window_size=(200, 200), batch=None):
        self.batch = pyglet.graphics.Batch()
        super().__init__(label, window_size=window_size, batch=self.batch)
        
        if label == "Paused":
            buttons = {0: ("Press [P] to unpause", CMD_CONTINUE)}
        else:
            buttons = {0: ("Got it! Start Game!", CMD_CONTINUE)}
        self.add_items(buttons, self.batch)
        self.reset()

        self.instructions_move = pyglet.text.Label(
            TEXT_INSTRUCTIONS_MOVE, 
            font_name=FONT_NAME,
            font_size=self.FONT_SIZE,
            x=window_size[0] // 2,
            y=window_size[1] * 2 // 3 - (self.TEXT_OFFSET),
            anchor_x='center',
            anchor_y='center',
            batch=self.batch)
                                      
        self.instructions_fire = pyglet.text.Label(
            TEXT_INSTRUCTIONS_FIRE, 
            font_name=FONT_NAME,
            font_size=self.FONT_SIZE,
            x=window_size[0] // 2,
            y=(window_size[1] * 2 // 3) - (self.TEXT_OFFSET * 2),
            anchor_x='center',
            anchor_y='center',
            batch=self.batch)

    def draw(self):
        # FIXME fix this draw thing. Something with hierarchy
        super().draw()
        self.batch.draw()


class InGameMenu(Menu):

    def __init__(self, window_size=(200, 200), batch=None):
        self.batch = pyglet.graphics.Batch()

        super().__init__('Quit game?', window_size=window_size, batch=self.batch)

        buttons = {0: ("Continue", CMD_CONTINUE),
                   1: ("Quit", CMD_QUIT)}
        self.add_items(buttons, self.batch)

        # self.items.append(MenuItem('New Game', 240, lambda: self.set_state("start")))
        # self.items.append(MenuItem('Instructions', 200, lambda: self.set_state("load_instructions")))
        # self.items.append(MenuItem('Options', 160, lambda: self.set_state("load_options")))
        # self.items.append(MenuItem('Quit', 120, lambda: self.set_state("quit")))
        self.reset()

    def draw(self):
        # FIXME fix this draw thing. Something with hierarchy
        super().draw()
        self.batch.draw()


class OptionsMenu(Menu):
    def __init__(self):
        super(OptionsMenu, self).__init__('Options')

        self.items.append(DifficultyMenuItem(280))

        def set_enable_sound(value):
            global enable_sound
            enable_sound = value

        self.items.append(ToggleMenuItem('Sound', enable_sound, 240,
                                         set_enable_sound))

        def set_enable_fullscreen(value):
            win.set_fullscreen(value, width=ARENA_WIDTH, height=ARENA_HEIGHT)

        self.items.append(ToggleMenuItem('Fullscreen', win.fullscreen, 200,
                                         set_enable_fullscreen))

        self.items.append(ToggleMenuItem('Vsync', win.vsync, 160,
                                         win.set_vsync))

        def set_show_fps(value):
            global show_fps
            show_fps = value

        self.items.append(ToggleMenuItem('FPS', show_fps, 120, set_show_fps))
        self.items.append(MenuItem('Ok', 60, begin_main_menu))
        self.reset()


        # batch=self.batch)

        # self.items.append(MenuItem('Continue Game (P)', 240, resume_game))
        # self.items.append(MenuItem('Main Menu', 200, end_game))
        # self.reset()
