import enum
from game import scoreboard
import pyglet

from pyglet.window import key

from . import projectile
from . import misc_sprites
from .train import Train, TrainSegment, TrainCarriage
from .resources.asset_definitions import *

from typing import Optional

PLAYER_SPEED = 1000 / 3.6
PLAYER_SCORE_THRESHOLD = 1.0

MOVE_NONE = 0
MOVE_LEFT = 1
MOVE_RIGHT = 2


class PlayerDir(enum.Enum):
    left = 0
    right = 1


class Player(pyglet.sprite.Sprite, key.KeyStateHandler):
    def __init__(
        self, x, y, batch, carriage_abilities, projectile_batch, 
        projectiles, scoreboard, health_effect_factory, 
        shield_effect_factory,
        train: Train, add_carriage, img=PLAYER_IMG):
        super(Player, self).__init__(img, x, y, batch=batch)
        self._sprite_imgs = {
            'left': img.get_transform(flip_x=True),
            'right': img
        }

        self._current_direction = PlayerDir.right
        self.carriage_abilities = carriage_abilities
        self.projectile_batch = projectile_batch
        self.dt_since_last_fired = 1000

        self.train_pos = 0
        self.move = MOVE_NONE
        self.train = train
        self.curr_segment = train.get_first_segment()
        self.health_effect_factory = health_effect_factory
        self.shield_effect_factory = shield_effect_factory

        self.health_timer = 100
        self.shield_timer = 100

        # Temporarily added here for 'survival points'
        self.scoreboard = scoreboard
        self.score_timer = 0

        self.last_important_score = 0
        self.add_carriage = add_carriage


    def move_left(self):
        self.move = MOVE_LEFT

    def move_right(self):
        self.move = MOVE_RIGHT

    def _move_to_left_if_invalid_segment(self):
        while (
            self.curr_segment is None
            or self.curr_segment not in self.train.get_segments()
        ):
            # try to stay in same spot if exists else move left
            if self.train.get_segment(self.train_pos) is None:
                self.train_pos -= 1
            self.curr_segment = self.train.get_segment(self.train_pos)

    def _move_to_left(self):
        if self.train_pos > 0:
            self.train_pos -= 1
        self.curr_segment = self.train.get_segment(self.train_pos)
        self._move_to_left_if_invalid_segment()

    def _move_to_right(self):
        if self.train_pos < len(self.train):
            self.train_pos += 1
        self.curr_segment = self.train.get_segment(self.train_pos)
        self._move_to_left_if_invalid_segment()

    def update(self, dt: float): #, start_bound_x: int, end_bound_x: int):
        # update based on train speed
        self.x += self.train.get_speed() * dt

        # update based on jump animation
        # TODO

        # discrete motion

        # segment was destroyed
        while (
            self.curr_segment is None
            or self.curr_segment not in self.train.get_segments()
        ):
            # try to stay in same spot if exists else move left
            if self.train.get_segment(self.train_pos) is None:
                self.train_pos -= 1
            self.curr_segment = self.train.get_segment(self.train_pos)

        if self.move in [MOVE_LEFT, MOVE_RIGHT]:
            if self.move == MOVE_LEFT:
                if self._current_direction == PlayerDir.right:
                    self.image = self._sprite_imgs['left']
                    self._current_direction = PlayerDir.left
                else:
                    self._move_to_left()
            elif self.move == MOVE_RIGHT:
                if self._current_direction == PlayerDir.left:
                    self.image = self._sprite_imgs['right']
                    self._current_direction = PlayerDir.right
                else:
                    self._move_to_right()
            self.move = MOVE_NONE
        # self.curr_segment = self.train.get_segment(self.train_pos)
        self.x = self.curr_segment.x 

        carriage_type: Optional[str] = (
            self.curr_segment.carriage_type
            if isinstance(self.curr_segment, TrainCarriage)
            else None
        )

        if self[key.SPACE]:
            carriage_ability = self.carriage_abilities.get(carriage_type)

            if carriage_ability is not None:
                if self.dt_since_last_fired > 0.4:
                    carriage_ability(
                        self.x, 
                        self.y, 
                        self.projectile_batch, 
                        (self.x, self.y + 100),  # target 
                        1.0,
                        self.curr_segment
                    )
                    self.dt_since_last_fired = 0.0

        self.dt_since_last_fired += dt

        if carriage_type == "health" and self.health_timer > 1.0:
            for carriage in self.train.segments:
                if (
                    carriage != self.curr_segment
                    and isinstance(carriage, TrainCarriage)
                ):
                    carriage.repair(1)
                    self.health_effect_factory.create(carriage.x, carriage.y)
            self.health_timer = 0.0
        self.health_timer += dt

        if self.shield_timer > 1.0:
            if carriage_type == "shield":
                for carriage in self.train.segments:
                    if isinstance(carriage, TrainCarriage):
                        if carriage == self.curr_segment:
                            carriage.disable_shield()
                        else:
                            carriage.enable_shield()
                            self.shield_effect_factory.create(carriage.x, carriage.y)
            else:
                for carriage in self.train.segments:
                    if isinstance(carriage, TrainCarriage):
                        carriage.disable_shield()
            self.shield_timer = 0.0

        self.shield_timer += dt

        if carriage_type == "loot":
            threshold = PLAYER_SCORE_THRESHOLD / 2
        else:
            threshold = PLAYER_SCORE_THRESHOLD

        if self.score_timer >= threshold:
            self.scoreboard.modify_score(1)
            self.score_timer = 0
        self.score_timer += dt


        if self.scoreboard.score != self.last_important_score and self.scoreboard.score % 10 == 0:
            self.add_carriage()
            self.last_important_score = self.scoreboard.score

    def get_position_x(self):
        return self.x

    def get_position_y(self):
        return self.y

    def set_position_y(self, position_y):
        self.y = position_y

    def set_position_x(self, position_x):
        self.x = position_x
