import math

import pyglet

from typing import List

from . import misc_sprites
from .resources.asset_definitions import CANNON_SFX, EXPLOSION_SFX


class Projectile(pyglet.sprite.Sprite):
    def __init__(self, img, x, y, batch, target, speed):
        super().__init__(img, x, y, batch=batch)
        self.target_x, self.target_y = target
        self.speed = speed
        self._done = False

    def update(self, dt: float):
        angle = math.atan2(self.target_y - self.y, self.target_x - self.x)
        self.x = self.x + math.cos(angle) * self.speed
        self.y = self.y + math.sin(angle) * self.speed

        if self.at_target():
            self._done = True
            
    def at_target(self):
        return (self.x - self.target_x)*(self.x - self.target_x) + (self.y - self.target_y)*(self.y - self.target_y) < 50

    @property
    def done(self):
        return self._done


class Dynamite(Projectile):
    def __init__(self, img, x, y, batch, target, speed):
        super().__init__(img, x, y, batch, target, speed)

    def update(self, dt: float):
        super().update(dt)

    @property
    def damage(self):
        return 5


class Rocket(Projectile):
    def __init__(self, img, x, y, batch, target, speed):
        super().__init__(img, x, y, batch, target, speed)

    def update(self, dt: float):
        angle = math.atan2(self.target_y - self.y, self.target_x - self.x)
        self.x = self.x + math.cos(angle) * self.speed
        self.y = self.y + math.sin(angle) * self.speed

        self.rotation = angle*180.0/math.pi - 90.0  # -90 cause image is facing up

        if self.at_target():
            self._done = True

    @property
    def damage(self):
        return 2


class CannonBall(Projectile):
    def __init__(self, img, x, y, batch, target, speed):
        super().__init__(img, x, y, batch, target, speed)

    @property
    def damage(self):
        return 20


class DynamiteFactory:
    def __init__(self, dynamite_image, projectiles):
        self.dynamite_image = dynamite_image
        self.dynamite_image.anchor_x = self.dynamite_image.width // 2
        self.dynamite_image.anchor_y = self.dynamite_image.height // 2
        self.projectiles = projectiles

    def create(self, x, y, batch, target, speed):
        dynamite = Dynamite(self.dynamite_image, x, y, batch, target, speed)
        dynamite.scale = 0.1
        self.projectiles.append(dynamite)
        return dynamite


class RocketFactory:
    def __init__(self, rocket_anim, projectiles):
        self.rocket_anim = rocket_anim
        self.rocket_anim.anchor_x = self.rocket_anim.width // 2
        self.rocket_anim.anchor_y = self.rocket_anim.height // 2
        self.projectiles = projectiles

    def create(self, x, y, batch, target, speed):
        rocket = Rocket(self.rocket_anim, x, y, batch, target, 10)
        self.projectiles.append(rocket)
        return rocket


class ExplosionFactory:
    def __init__(self, explosion_animation, audio_player):
        self.explosion_animation = explosion_animation
        self.audio_player = audio_player
    
    def create(self, x, y, batch):
        explosion = misc_sprites.EffectSprite(self.explosion_animation, x, y, batch=batch)
        explosion.scale = 5
        EXPLOSION_SFX.play()
        # Use audio player for volume, but looks like we need audio groups.
        # Or dynamically create a player per sound if we want to control effect and volume.
        # self.audio_player.queue(EXPLOSION_SFX)
        return explosion


class CannonBallFactory:
    def __init__(self, cannon_ball_image, projectile_batch, projectiles, audio_player):
        self.cannon_ball_image = cannon_ball_image
        self.cannon_ball_image.anchor_x = self.cannon_ball_image.width // 2
        self.cannon_ball_image.anchor_y = self.cannon_ball_image.height // 2
        self.projectile_batch = projectile_batch
        self.projectiles = projectiles
        self.audio_player = audio_player

    def create(self, x, y, target, speed):
        cannon_ball = CannonBall(
            self.cannon_ball_image, x, y, self.projectile_batch, target, speed
        )
        self.projectiles.append(cannon_ball)
        CANNON_SFX.play()
        # self.audio_player.queue(CANNON_SFX)
        return cannon_ball


def update(
    projectiles: List[Projectile], 
    explosions: List[misc_sprites.EffectSprite],
    damageables: List,
    explosion_factory: ExplosionFactory,
    dt: float, 
    projectiles_batch: pyglet.graphics.Batch
):
    for projectile in projectiles:
        projectile.update(dt)

        if projectile.done:
            explosions.append(explosion_factory.create(projectile.x, projectile.y, projectiles_batch))


            for damageable in damageables:
                if damageable.dead:
                    continue

                dist_sq = (
                    (damageable.y - projectile.y)*(damageable.y - projectile.y)
                    + (damageable.x - projectile.x)*(damageable.x - projectile.x)
                )
                radius = damageable.radius()
                if dist_sq < radius*radius:
                    damageable.damage(projectile.damage)

                    if damageable.dead:
                        explosion = explosion_factory.create(damageable.x, damageable.y, projectiles_batch)
                        explosion.scale = 20
                        explosions.append(explosion)

            projectile.delete()
    
    projectiles[:] = (projectile for projectile in projectiles if not projectile.done)
    damageables[:] = (damageable for damageable in damageables if not damageable.dead)
    explosions[:] = (explosion for explosion in explosions if not explosion.done)