# PyWeek 32 2021
# Written by scott, jttofu and jnruan

import math
import random

import pyglet
from pyglet.gl import *
from pyglet.window import key

from . import train
from . import ship
from . import player
from . import projectile
from . import scoreboard
from . import misc_sprites
from . import overlay

from .resources.asset_definitions import *


class Game:
    # WINDOW_WIDTH = pyglet.canvas.Display().get_default_screen().width
    # WINDOW_HEIGHT = pyglet.canvas.Display().get_default_screen().height
    WINDOW_WIDTH = 1600
    WINDOW_HEIGHT = 900
    WINDOW_MIN_SIZE = (640, 480)
    GAME_TITLE = 'Neverending cowboys vs pirates'

    def __init__(self):
        self.configure_window()

        self.camera_x = 0

        self.paused = False
        self.game_over = False

        self.audio_player = pyglet.media.Player()
        self.audio_player.volume = 0.5
        self.audio_player.play()  # Start playing anything that gets queued.

        self.music_player = pyglet.media.Player()
        self.music_player.volume = 0.5

        # --------------------------------------------------------------------------
        # Global game state vars
        # --------------------------------------------------------------------------
        self.train_carriage_gap = 20.0
        self.train_start_x = 255.0
        self.train_start_y = 240.0
        self.train_scale = 0.5
        self.train_speed = -1000.0 / 3.6  # m/s

        self.player_start_offset_x = 100

        self.enemies_offset = (100.0 + self.train_start_x,
                               270.0)
        self.enemies_speed = self.train_speed

        self.init_sprites()

    def configure_window(self):
        self.window = pyglet.window.Window(
            self.WINDOW_WIDTH,
            self.WINDOW_HEIGHT,
            caption=self.GAME_TITLE,
            resizable=True,
            vsync=True
        )

        self.window.set_minimum_size(*self.WINDOW_MIN_SIZE)

        fps = pyglet.window.FPSDisplay(self.window)
        # fps = pyglet.clock.ClockDisplay()

        self.overlay_batch = pyglet.graphics.Batch()
        self.overlay_manager = overlay.OverlayManager(window=self.window, batch=self.overlay_batch)
        self.overlay_manager.set_overlay(overlay.OVERLAY_SPLASH)

        @self.window.event
        def on_resize(width, height):
            print('The window was resized to %dx%d' % (width, height))
            aspect_ratio = self.WINDOW_HEIGHT / self.WINDOW_WIDTH
            new_height = width * aspect_ratio
            glMatrixMode(gl.GL_PROJECTION)
            glLoadIdentity()
            glViewport(0, 0, width, int(new_height))
            gluOrtho2D(0, self.WINDOW_WIDTH, 0, self.WINDOW_HEIGHT)
            glMatrixMode(gl.GL_MODELVIEW)

        self.window.on_resize = on_resize

        @self.window.event
        def on_key_press(symbol, modifiers):
            # Overrides default Escape key behaviour
            if symbol == key.ESCAPE:
                self.overlay_manager.set_overlay(overlay.OVERLAY_INGAME_SETTINGS)
                # self.window.close()
            elif symbol == key.LEFT:
                self.player1.move_left()
            elif symbol == key.RIGHT:
                self.player1.move_right()
            elif symbol == key.P:
                self.paused = not self.paused
                if self.paused:
                    self.overlay_manager.set_overlay(overlay.OVERLAY_PAUSED)
                else:
                    self.overlay_manager.set_overlay(overlay.OVERLAY_NONE)
            return pyglet.event.EVENT_HANDLED

        @self.window.event
        def on_draw():
            glClearColor(0.5, 0.5, 0.8, 1.0)
            self.window.clear()
            glColor3f(1.0, 1.0, 1.0)

            glLoadIdentity()
            glTranslatef(self.camera_x, 0.0, 0.0)

            self.background_sprite.draw()

            self.enemies.batch_draw()

            for railway_sprite in self.railway_sprites:
                railway_sprite.draw()

            self.train_batch.draw()

            self.projectile_batch.draw()

            self.enemies.hud_batch_draw()
            self.train_hud_batch.draw()

            glLoadIdentity()
            fps.draw()

            self.game_hud_batch.draw()
            self.overlay_manager.draw()



    def init_sprites(self):
        # setup projectiles
        self.projectiles = []
        self.explosions = []
        self.damageables = []
        self.effects = []
        # self.enemies = []

        # HUD
        self.game_hud_batch = pyglet.graphics.Batch()
        self.scoreboard = scoreboard.ScoreBoard(self.game_hud_batch, 50, 100, 0, 1)

        # create background
        self.background_sprite = pyglet.sprite.Sprite(BACKGROUND_IMG, -1920 * 2, 0)
        self.railway_sprites = [
            pyglet.sprite.Sprite(RAILWAY_IMG, RAILWAY_IMG.width * i, self.train_start_y - 60)
            for i in [-4, -3, -2, -1, 0, 1]
        ]

        # create train
        train_sequence = [
            ('engine', '', 100),
            ('carriage', 'dynamite', 100),
            #('carriage', 'rocket', 100),
            #('carriage', 'loot', 100),
            #('carriage', 'shield', 100),
            #('carriage', 'health', 100)
        ]

        self.train_batch = pyglet.graphics.Batch()
        self.train_hud_batch = pyglet.graphics.Batch()

        self.train_carriage_factory = train.TrainCarriageFactory(scale=self.train_scale, damageables=self.damageables)
        self.train_factory = train.TrainFactory(train_sequence,
                                                self.train_carriage_factory,
                                                self.train_carriage_gap,
                                                self.train_scale,
                                                damageables=self.damageables)
        self.train1 = self.train_factory.create(train_start_x=self.train_start_x,
                                                train_start_y=self.train_start_y,
                                                train_speed=self.train_speed,
                                                batch=self.train_batch,
                                                hud_batch=self.train_hud_batch)

        self.projectile_batch = pyglet.graphics.Batch()
        self.dynamite_factory = projectile.DynamiteFactory(DYNAMITE_IMG, self.projectiles)
        self.rocket_factory = projectile.RocketFactory(ROCKET_IMG, self.projectiles)
        self.explosion_factory = projectile.ExplosionFactory(EXPLOSION_ANIMATION, self.audio_player)
        self.health_effect_factory = misc_sprites.HealthEffectFactory(HEALTH_ANIMATION, self.projectile_batch, self.effects)
        self.shield_effect_factory = misc_sprites.ShieldEffectFactory(SHIELD_ANIMATION, self.projectile_batch, self.effects)

        # create pirate ship enemies
        self.enemies = ship.Enemies(start_pos=self.enemies_offset,
                                    speed=self.enemies_speed,
                                    damageables=self.damageables,
                                    projectile_batch=self.projectile_batch,
                                    projectiles=self.projectiles,
                                    scoreboard=self.scoreboard,
                                    audio_player=self.audio_player)

        # setup player
        player_start_x = self.train1.start_x + self.player_start_offset_x
        player_start_y = self.train1.start_y + self.train1.height  # self.train_carriage_image.height * self.train_scale

        carriage_abilities = {
            "dynamite": lambda x, y, projectile_batch, target, speed, current_segment: self.dynamite_factory.create(x, y, projectile_batch, target, speed),
            "rocket":  lambda x, y, projectile_batch, target, speed, current_segment: self.rocket_factory.create(x, y, projectile_batch, target, speed),
        }

        def add_carriage():
            carriage_types = train.TrainCarriage.carriage_types
            carriage_type = carriage_types[random.randint(0, len(carriage_types)-1)]
            segment = self.train_carriage_factory.create(
                carriage_type=carriage_type,
                x=self.train1.end_x,
                y=self.train1.start_y,
                batch=self.train_batch,
                hud_batch=self.train_hud_batch,
                hp=100
            )
            self.train1.add_segment(segment)
            self.train1.fix_segment_positions(len(self.train1)-2)

        self.player1 = player.Player(player_start_x, player_start_y,
                                     self.train_batch,
                                     carriage_abilities,
                                     self.projectile_batch,
                                     self.projectiles,
                                     self.scoreboard,
                                     self.health_effect_factory,
                                     self.shield_effect_factory,
                                     train=self.train1,
                                     add_carriage=add_carriage)

        self.window.push_handlers(self.player1)

    def get_halted(self):
        return self.overlay_manager.is_displaying() or self.paused # or self.game_over

    # --------------------------------------------------------------------------
    # Game update
    # --------------------------------------------------------------------------
    def update(self, dt: float):
        if len(self.train1) <= 1:
            self.game_over = True

        if self.get_halted():
            return

        self.camera_x -= self.train_speed * dt

        if self.camera_x > BACKGROUND_IMG.width // 2:
            player_offset_x = self.player1.get_position_x() - self.train1.segments[0].x + self.train_start_x
            self.camera_x = self.camera_x - BACKGROUND_IMG.width // 2

            self.train1.set_position_x(self.camera_x + self.train_start_x)
            self.player1.set_position_x(self.camera_x + player_offset_x)

            self.enemies.set_position_x(self.camera_x)

        self.train1.update(dt)

        if not self.game_over:
            self.player1.update(dt)  # train_start, train_end)
            self.enemies.update(dt, pos_x=-self.camera_x)
        else:
            self.train_speed -= self.train_speed*0.01
            self.train1.train_speed = self.train_speed

            if abs(self.train_speed) < 5:
                self.overlay_manager.set_overlay(overlay.OVERLAY_GAME_OVER)
                self.game_over = False
                self.train_speed = 10

        projectile.update(self.projectiles, self.explosions, self.damageables,
                          self.explosion_factory, dt, self.projectile_batch)



# --------------------------------------------------------------------------
# Main entry point
# --------------------------------------------------------------------------
def run():
    game = Game()

    pyglet.clock.schedule_interval(game.update, 1 / 60.)

    glEnable(GL_BLEND)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    glDisable(GL_DEPTH_TEST)

    pyglet.app.run()
