import math
import random

import pyglet

from typing import List, Tuple

from . import misc_sprites

from .resources.asset_definitions import *
from . import projectile


class Enemies:
    SPAWN_TIMER_BASE = 5.0

    def __init__(self, start_pos: Tuple, speed: float, damageables: List, projectile_batch, projectiles, scoreboard, audio_player):
        self.ship_start_x = start_pos[0] + PIRATE_SHIP_IMG.width // 2
        self.ship_start_y = start_pos[1] + PIRATE_SHIP_IMG.height // 2
        self.ship_speed = speed
        self.scoreboard = scoreboard

        self.enemies = []

        self.difficulty = 1
        self.ships_spawned = 1
        self.spawn_timer = 0
        self.spawn_active_threshold = self.SPAWN_TIMER_BASE

        self.ship_batch = pyglet.graphics.Batch()
        self.ship_hud_batch = pyglet.graphics.Batch()

        self.ship_group = pyglet.graphics.OrderedGroup(0)
        self.ship_armaments_group = pyglet.graphics.OrderedGroup(1)

        self.cannon_ball_factory = projectile.CannonBallFactory(CANNON_BALL_IMG, projectile_batch, projectiles, audio_player)
        self.cannon_factory = CannonFactory(CANNON_IMG, self.cannon_ball_factory)

        self.pirate_ship_factory = PirateShipFactory(damageables)

        self.pirate_ship = self.pirate_ship_factory.create(self.ship_start_x,
                                                          800,
                                                           self.ship_start_y,
                                                           self.ship_speed,
                                                           self.ship_batch,
                                                           self.ship_group,
                                                           self.ship_armaments_group,
                                                           self.ship_hud_batch,
                                                           100,
                                                           self.cannon_factory)

        self.enemies.append(self.pirate_ship)

    def batch_draw(self):
        self.ship_batch.draw()

    def hud_batch_draw(self):
        self.ship_hud_batch.draw()

    def set_position_x(self, x):
        for enemy in self.enemies:
            enemy.x = x + enemy.offset_x

    def update(self, dt, pos_x):
        if len(self.enemies) == 0:
            self.difficulty += 1
            self.scoreboard.level = self.difficulty
            self.ships_spawned = 0

        # First ship should spawn on level start
        if self.ships_spawned == 0 or (self.ships_spawned != self.difficulty and self.spawn_timer >= self.spawn_active_threshold):
            self.enemies.append(
                self.pirate_ship_factory.create(
                    pos_x,
                    200*self.ships_spawned + 800,
                    self.ship_start_y,
                    self.ship_speed,
                    self.ship_batch,
                    self.ship_group,
                    self.ship_armaments_group,
                    self.ship_hud_batch,
                    100,
                    self.cannon_factory
                )
            )
            self.ships_spawned += 1
            self.spawn_timer = 0
            # HACK: Add some random time [float, float] to avoid spawning at 'too similar' times - otherwise cannon SFX sounds like punishment
            self.spawn_active_threshold = self.SPAWN_TIMER_BASE + random.uniform(0.1, 0.9) 
        else:
            self.spawn_timer += dt

        for enemy in self.enemies:
            if not enemy.dead:
                enemy.update(dt, pos_x)
            else:
                self.scoreboard.modify_score(enemy.initial_hp)
                enemy.destroy()

        self.enemies[:] = (enemy for enemy in self.enemies if not enemy.dead)


class Cannon(pyglet.sprite.Sprite):
    def __init__(self, img, ship_x, offset_x, y, batch, group, cannon_ball_factory):
        super().__init__(img, ship_x + offset_x, y, batch=batch, group=group)
        self.offset_x = offset_x
        self.cannon_ball_factory = cannon_ball_factory
        self.time_since_last_fired = 1000

    def update(self, dt: float, ship_x: float, target_x):
        self.x = ship_x + self.offset_x

        self.time_since_last_fired += dt

        if self.time_since_last_fired > 2.0:
            self.time_since_last_fired = 0.0
            self.cannon_ball_factory.create(
                self.x, self.y,
                (self.x - 300 + target_x, self.y - 75),
                10.0
            )


class PirateShip(pyglet.sprite.Sprite):
    def __init__(self, img, x, offset_x, y, ship_speed: float, batch, group, hud_batch, initial_hp):
        super().__init__(img, x, y, batch=batch, group=group),

        self.ship_speed = 0
        self.ship_accel = 0.0
        self.hp = initial_hp
        self.initial_hp = initial_hp
        self.offset_x = offset_x

        self.hp_bar = misc_sprites.HPBar(
            hud_batch,
            self.x - self.image.width // 2,
            self.y - self.image.height // 2,
            self.image.width,
            5,
            600,
            10,
            initial_hp
        )

        self.cannons = []

    def update(self, dt: float, train_x):
        if abs(self.x - train_x) > 1000.0:
            self.ship_accel = -10.0
        else:
            self.ship_accel = 10.0

        self.ship_speed += self.ship_accel * dt
        self.offset_x += self.ship_speed * dt
        
        self.x = train_x + self.offset_x
        # self.x = train_x + self.offset_x

        for i, cannon in enumerate(self.cannons):
            cannon.update(dt, self.x, i * 100)

        self.hp_bar.update(dt, self.x - self.image.width // 2, self.initial_hp, self.hp)

    def set_position_x(self, position_x):
        self.x = position_x

    def damage(self, delta_hp):
        self.hp -= delta_hp

    @property
    def dead(self):
        return self.hp < 0.0

    def destroy(self):
        for cannon in self.cannons:
            cannon.delete()
        self.delete()

    def add_cannon(self, cannon: Cannon):
        self.cannons.append(cannon)

    def radius(self):
        return self.image.width // 2


class CannonFactory:
    def __init__(self, canon_image, cannon_ball_factory):
        self.cannon_image = canon_image
        self.cannon_image.anchor_x = self.cannon_image.width // 2
        self.cannon_image.anchor_y = self.cannon_image.height // 2
        self.cannon_ball_factory = cannon_ball_factory

    def create(self, x, y, batch, group):
        cannon = Cannon(self.cannon_image, 0, x, y, batch, group, self.cannon_ball_factory)
        return cannon


class PirateShipFactory:
    def __init__(self, damageables: List):
        self.damageables = damageables

    def create(
        self, x, offset_x, y, ship_speed: float, batch, ship_group, ship_armaments_group, hud_batch, initial_hp, cannon_factory,
        img=PIRATE_SHIP_IMG
    ) -> PirateShip:
        pirate_ship = PirateShip(img, x, offset_x, y, ship_speed, batch, ship_group, hud_batch, initial_hp)

        self.damageables.append(pirate_ship)

        for cannon_x in [300, 450, 600]:
            pirate_ship.add_cannon(
                cannon_factory.create(
                    cannon_x - 769 // 2,
                    pirate_ship.y + 50 - 746 // 2,
                    batch,
                    ship_armaments_group
                )
            )

        return pirate_ship
