import math
import random
import itertools

import pyglet

from typing import List, Tuple, Optional

from . import misc_sprites

from .resources.asset_definitions import *


class Wheel(pyglet.sprite.Sprite):
    def __init__(self, x, y, batch=None):
        img = WHEEL_IMG
        img.anchor_x = img.width // 2
        img.anchor_y = img.height // 2
        super().__init__(img, x, y, batch=batch)

        self.rotation = random.random() * 360

    def update(self, dt: float, train_speed: float):
        self.x += train_speed * dt

        rotation = self.rotation + train_speed * dt * 0.05 * 45.0
        if rotation < -180.0:
            rotation += 360.0

        self.rotation = rotation


class TrainSegment(pyglet.sprite.Sprite):
    def __init__(self, img, x, y, batch):
        super().__init__(img, x, y, batch=batch)
        self.wheels = []
        self.wheel_offset_x = []

    def add_wheel(self, wheel: Wheel, wheel_offset_x: float):
        self.wheels.append(wheel)
        self.wheel_offset_x.append(wheel_offset_x)

    def update(self, dt: float, train_speed: float):
        self.x += train_speed * dt

        for wheel in self.wheels:
            wheel.update(dt, train_speed)

    def set_position_x(self, position_x: float):
        self.x = position_x

        for wheel, wheel_offset_x in zip(self.wheels, self.wheel_offset_x):
            wheel.x = -self.image.width // 2 * self.scale + position_x + wheel_offset_x


class TrainEngine(TrainSegment):
    def __init__(self, x, y, batch=None):
        super().__init__(ENGINE_IMG, x, y, batch=batch)

    @property
    def dead(self):
        return False


class TrainCarriage(TrainSegment):
    carriage_types = [
        "dynamite",
        "rocket",
        "loot",
        "shield",
        "health"
    ]

    def __init__(self, x, y, scale, batch, hud_batch, initial_hp, img=CARRIAGE_IMG):
        super().__init__(img, x, y, batch=batch)
        self.scale = scale
        self.initial_hp = initial_hp
        self.hp = initial_hp
        self.hp_bar = misc_sprites.HPBar(hud_batch,
                                         self.x - self.image.width // 2 * self.scale,
                                         self.y - self.image.height // 2 * self.scale,
                                         self.image.width * self.scale,
                                         5,
                                         -35,
                                         10,
                                         initial_hp
                                         )
        self.shield = True

    def update(self, dt: float, train_speed: float):
        super(TrainCarriage, self).update(dt, train_speed)
        self.hp_bar.update(dt, self.x - self.image.width // 2 * self.scale, self.initial_hp, self.hp)

    def damage(self, amount):
        if self.shield:
            amount = 0.1 * amount

        if self.hp >= 0.0:
            self.hp -= amount

        if self.hp < 0.0:
            self.hp = -1.0

    def repair(self, amount):
        if self.hp < self.initial_hp:
            self.hp += 20.0 if self.initial_hp - self.hp > 20.0 else self.initial_hp - self.hp

    def enable_shield(self):
        self.shield = True

    def disable_shield(self):
        self.shield = False

    @property
    def dead(self):
        return self.hp < 0.0

    def radius(self):
        return self.image.width // 2


class TrainCarriageDynamite(TrainCarriage):
    carriage_type = "dynamite"

    def __init__(self, x, y, scale, batch, hud_batch, initial_hp):
        super().__init__(x, y, scale, batch, hud_batch, initial_hp, CARRIAGE_DYNAMITE_IMG)


class TrainCarriageRocket(TrainCarriage):
    carriage_type = "rocket"

    def __init__(self, x, y, scale, batch, hud_batch, initial_hp):
        super().__init__(x, y, scale, batch, hud_batch, initial_hp, CARRIAGE_ROCKET_IMG)


class TrainCarriageLoot(TrainCarriage):
    carriage_type = "loot"
    
    def __init__(self, x, y, scale, batch, hud_batch, initial_hp):
        super().__init__(x, y, scale, batch, hud_batch, initial_hp, CARRIAGE_LOOT_IMG)


class TrainCarriageShield(TrainCarriage):
    carriage_type = "shield"

    def __init__(self, x, y, scale, batch, hud_batch, initial_hp):
        super().__init__(x, y, scale, batch, hud_batch, initial_hp, CARRIAGE_SHIELD_IMG)


class TrainCarriageHealth(TrainCarriage):
    carriage_type = "health"

    def __init__(self, x, y, scale, batch, hud_batch, initial_hp):
        super().__init__(x, y, scale, batch, hud_batch, initial_hp, CARRIAGE_HEALTH_IMG)


class Train:
    segments: List[TrainSegment]
    train_start_x: float
    train_start_y: float
    train_end_x: float
    train_end_y: float
    gap_x: float

    def __init__(self, train_start_x, train_start_y, train_speed: float, gap_x: float, train_height: int = 0):
        self.segments = []
        self.train_start_x = train_start_x
        self.train_start_y = train_start_y
        self.train_end_x = train_start_x
        self.train_end_y = train_start_y
        self.train_speed = train_speed
        self.gap_x = gap_x

        self.train_height = train_height

    def get_speed(self):
        return self.train_speed

    @property
    def start_x(self) -> float:
        return self.train_start_x

    @property
    def start_y(self) -> float:
        return self.train_start_y

    @property
    def end_x(self) -> float:
        return self.train_end_x

    @property
    def end_y(self) -> float:
        return self.train_end_y

    @property
    def height(self):
        return self.train_height

    def add_segment(self, train_segment: TrainSegment):
        self.segments.append(train_segment)
        self.train_end_x += self.gap_x + train_segment.width

    def remove_segment(self, train_segment: TrainSegment):
        index = self.get_segment_index(train_segment)
        if index is None:
            return

        self.segments.remove(train_segment)
        self.fix_segment_positions(index-1)
        self.train_end_x -= self.gap_x + train_segment.width

    def fix_segment_positions(self, index: int):
        prior_segment = self.get_segment(index)

        if prior_segment is None:
            return

        # shift remaining segments forward
        for segment in self.segments[index+1:]:
            segment.set_position_x(prior_segment.x + self.gap_x + prior_segment.width // 2 + segment.width // 2)
            prior_segment = segment
            

    def get_first_segment(self):
        return self.segments[0]

    def get_last_segment(self):
        return self.segments[-1]

    def __len__(self):
        return len(self.segments)

    def get_start(self):
        return self.get_first_segment().wheels[0].x

    def get_end(self):
        return self.get_last_segment().wheels[-1].x

    def get_segment(self, i) -> Optional[TrainSegment]:
        try:
            return self.segments[i]
        except IndexError:
            return None

    def get_segments(self):
        # copy from self.segments
        segment_list = [segment for segment in self.segments]
        return segment_list

    def get_segment_index(self, segment: TrainSegment):
        try:
            index = self.segments.index(segment)
        except ValueError:
            index = None
        return index

    def update(self, dt: float):
        # remove dead segments
        dead_segments = [segment for segment in self.segments if segment.dead]

        for dead_segment in dead_segments:
            self.remove_segment(dead_segment)

        # self.segments[:] = (segment for segment in self.segments if not segment.dead)

        for segment in self.segments:
            segment.update(dt, self.train_speed)

    def set_position_x(self, start_position_x: float):
        position_x = start_position_x
        for segment in self.segments:
            segment.set_position_x(position_x)
            position_x += self.gap_x + segment.width


class TrainEngineFactory:
    def __init__(self, scale: float):  # train_engine_image, train_wheel_image, scale: float):
        # self.train_engine_image = train_engine_image
        # self.train_wheel_image = train_wheel_image
        self.scale = scale
        self.train_height = ENGINE_IMG.height * scale

    @property
    def height(self):
        return self.train_height

    def create(self, x, y, batch=None) -> TrainEngine:
        train_engine = TrainEngine(x, y, batch)
        train_engine.scale = self.scale

        wheel_xs = [95.0 * self.scale, 275.0 * self.scale, 340.0 * self.scale]
        wheel_y = 20 * self.scale
        wheel_scales = [1.5 * self.scale, 2.0 * self.scale, 2.0 * self.scale]

        for wheel_x, scale in zip(wheel_xs, wheel_scales):
            wheel = Wheel(x + wheel_x - 110,
                          y + wheel_y - 40,
                          batch
                          )
            wheel.scale = scale
            # wheel.update_anchor(self.train_wheel_image, scale)
            train_engine.add_wheel(wheel, wheel_x)

        return train_engine


class TrainCarriageFactory:

    def __init__(self, scale: float, damageables: List):
        # self.train_wheel_image = train_wheel_image

        self.scale = scale
        self.damageables = damageables
        self.ctors = {
            "dynamite": TrainCarriageDynamite,
            "rocket": TrainCarriageRocket,
            "loot": TrainCarriageLoot,
            "shield": TrainCarriageShield,
            "health": TrainCarriageHealth
        }

    def create(
        self, carriage_type: str, x: float, y: float, batch, hud_batch, hp: float
    ) -> TrainCarriage:

        ctor = self.ctors[carriage_type]

        train_carriage = ctor(x, y, self.scale, batch, hud_batch, hp)
        self.damageables.append(train_carriage)

        wheel_xs = [50.0 * self.scale, 210.0 * self.scale]
        wheel_y = 20 * self.scale
        wheel_scales = [2.0 * self.scale, 2.0 * self.scale]

        for wheel_x, scale in zip(wheel_xs, wheel_scales):
            wheel = Wheel(x + wheel_x - 65,
                          y + wheel_y - 40,
                          batch
                          )
            wheel.scale = scale
            train_carriage.add_wheel(wheel, wheel_x)

        return train_carriage


class TrainFactory:
    train_engine_factory: TrainEngineFactory
    train_carriage_factory: TrainCarriageFactory

    # def __init__(self, train_engine_factory, train_carriage_factory, gap_x: float, scale: float):
    # self.train_engine_factory = train_engine_factory
    # self.train_carriage_factory = train_carriage_factory
    def __init__(
        self, segments: List[Tuple[str, str, int]], train_carriage_factory, gap_x: float, scale: float, damageables: List
    ):
        self.train_carriage_factory = train_carriage_factory
        self.train_engine_factory = TrainEngineFactory(scale=scale)
        self.segments = segments
        self.gap_x = gap_x
        self.scale = scale

    def create(self, train_start_x, train_start_y, train_speed: float, batch=None, hud_batch=None):
        train = Train(train_start_x, train_start_y, train_speed, self.gap_x)

        for segment_type, carriage_type, hp in self.segments:
            if segment_type == 'engine':
                segment = self.train_engine_factory.create(
                    x=train.end_x, y=train.end_y, batch=batch
                )
            elif segment_type == 'carriage':
                segment = self.train_carriage_factory.create(
                    carriage_type=carriage_type, 
                    x=train.end_x, 
                    y=train.start_y, 
                    batch=batch,
                    hud_batch=hud_batch,
                    hp=hp
                )
            else:
                raise NotImplementedError
            train.add_segment(segment)

        # set train height for player
        train.train_height = CARRIAGE_IMG.height * self.scale

        return train
